CON
  _clkfreq = 80_000_000
  _clkmode = xtal1 + pll16x

  'SDCard Interface
  _cardDataOutPin = 16
  _cardClockPin = 21
  _cardDataInPin = 20
  _cardChipSelectPin = 19

  'W5100 Interface
  W5100_DATA0 = 0
  W5100_DATA1 = 1
  W5100_DATA2 = 2
  W5100_DATA3 = 3
  W5100_DATA4 = 4
  W5100_DATA5 = 5
  W5100_DATA6 = 6
  W5100_DATA7 = 7
  W5100_ADDR0 = 8
  W5100_ADDR1 = 9
  W5100_WR    = 10
  W5100_RD    = 11
  W5100_CS    = 12
  W5100_INT   = 13
  W5100_RST   = 14
  W5100_SEN   = 15

  XLED         = 23   ' = extra user led Yellow (D1)

  CR            = 13
  LF            = 10
  QUOTE         = 34
  
  bytebuffersize = 2048
  maxPacketSendSize = 2032
  listenPort = 80


    
OBJ
  SDCard        : "SD2.0_FATEngine"
  DHCPClient    : "DHCP"
  W5100         : "Brilldea_W5100_Indirect_Driver_Ver006.spin"
  PST           : "Parallax Serial Terminal"
  STR           : "STREngine"

VAR
  byte MAC_Address[6]
  byte IP[4]
  byte SubnetMask[4]
  byte GatewayIP[4]
  byte DNS_Server[4]

  long DHCP_LeaseTime           ' duration of lease in seconds
  
  byte HostName[64]
  byte DomainName[64]          ' you may need to make this larger if you have a very long domain name

  byte destIP[4]
  byte data[bytebuffersize]

  byte WantDHCP, ReadMACFromFile
  byte Counter, EditorFileName[13]

  byte led_stat
  long  PageCount


PUB main | index, fileSize, token, packetSize, fileName[64] ' 64 longs = 256 bytes
  PST.Start(115200)
  SDCard.FATEngineStart(_cardDataOutPin, _cardClockPin, _cardDataInPin, _cardChipSelectPin, 0, 0)

  dira[XLED] := 1
  'ledje := 1
  '
  ' Network Settings
  '
  WantDHCP          := 1         '1=yes  0=no
  ReadMACFromFile   := 1         '1=yes  0=no       'if yes, will read from "MAC_ADDR" file 6 Bytes in Hex Form
  '
  ' If we don't want DHCP (WantDHCP:=0) these values will be used, otherwise they will be overwritten by the DHCPClient
  '
  ' IP Address
  IP[0]          := 192
  IP[1]          := 168
  IP[2]          := 15
  IP[3]          := 99
  ' Subnet Mask
  SubnetMask[0]  := 255
  SubnetMask[1]  := 255
  SubnetMask[2]  := 255
  SubnetMask[3]  := 0
  ' Gateway
  GatewayIP[0]   := 192 
  GatewayIP[1]   := 168 
  GatewayIP[2]   := 15 
  GatewayIP[3]   := 1
  ' DNS Server
  DNS_Server[0]   := 208 
  DNS_Server[1]   := 67   
  DNS_Server[2]   := 222   
  DNS_Server[3]   := 222  
  ' Destination IP address - can be left zeros, the demo echoes to computer that sent the packet
  destIP[0]      := 0
  destIP[1]      := 0
  destIP[2]      := 0
  destIP[3]      := 0
  '
  ' If we don't want to Read the MAC Address from a file (ReadMACFromFile:=0) these values will be used, otherwise they will be overwritten later
  '
  ' MAC Address
  MAC_Address[0] := $00
  MAC_Address[1] := $00
  MAC_Address[2] := $00 
  MAC_Address[3] := $00 
  MAC_Address[4] := $00 
  MAC_Address[5] := $00    

  ' delay to allow time to enable PST
  PauseMSec(2000)
  
  PST.Home
  PST.Clear
  PST.Str(string("HTTP Demo", PST#NL, PST#NL))
  PST.Str(string("Mounting SD card.", PST#NL, PST#NL))
  result := SDCard.mountPartition(0, (" "))
  'if STRSIZE(result) > 0
    'PST.Str(result)
    'abort    ''Was getting "Bp" then abort...commented this out and all works well.

  If ReadMACFromFile == 1                            ' If we want to read the MAC Address from "MAC_ADDR" on the SD Card
    PST.Str(string("Getting MAC Address from file."))
    SDCard.openFile(String("MAC_ADDR"), "r")
    repeat Counter from 0 to 6
      SDCard.readData(@data[0], 2)
      data[2]:=0
      MAC_Address[Counter] := STR.hexadecimalToNumber(@data[0])
    SDCard.closeFile
  else
    PST.Str(string("Manual MAC Address Chosen."))
  PST.Str(string(PST#NL, PST#NL))  


  If WantDHCP == 1                                   ' If we want DHCP try to grab an IP until we do
    DHCPClient.Start
    DHCPClient.SetMAC_Address(@MAC_Address[0])
    PST.Str(string("Doing DHCP."))
    PST.RxFlush
    repeat
    
      result := DHCPClient.DoDHCP(0, $42424242)
      DHCPClient.GetIP(@IP[0])
      DHCPClient.GetDNSServer(@DNS_Server[0])
      DHCPClient.GetGatewayIP(@GatewayIP[0])
      DHCPClient.GetSubnetMask(@SubnetMask[0])
      bytemove(@DomainName[0], DHCPClient.GetDomainName, 64)
      bytemove(@HostName[0], DHCPClient.GetHostName, 64)
       
      if result <> DHCPCLIENT#SUCCESS      'didn't happen, lets try again....print a "." char on PST to let you know
        
        PST.Str(string("."))

    WHile result <> DHCPCLIENT#SUCCESS

    'done with DHCP so stop it to release the cog it used for W5100        
    DHCPClient.Stop
    PST.Str(string("Done", PST#NL, PST#NL))
  else
    PST.Str(string("Manual IP Settings Chosen.", PST#NL, PST#NL))  
                                               
  PST.Str(string("IP Address:  "))
  PST.Dec(IP[0])
  PST.Str(string("."))
  PST.Dec(IP[1])
  PST.Str(string("."))
  PST.Dec(IP[2])
  PST.Str(string("."))
  PST.Dec(IP[3])
  PST.Str(string(PST#NL))

  PST.Str(string("Subnet Mask: "))
  PST.Dec(SubnetMask[0])
  PST.Str(string("."))
  PST.Dec(SubnetMask[1])
  PST.Str(string("."))
  PST.Dec(SubnetMask[2])
  PST.Str(string("."))
  PST.Dec(SubnetMask[3])
  PST.Str(string(PST#NL))    

  PST.Str(string("Gateway:     "))
  PST.Dec(GatewayIP[0])
  PST.Str(string("."))
  PST.Dec(GatewayIP[1])
  PST.Str(string("."))
  PST.Dec(GatewayIP[2])
  PST.Str(string("."))
  PST.Dec(GatewayIP[3])
  PST.Str(string(PST#NL))    

  PST.Str(string("DNS Server:  "))
  PST.Dec(DNS_Server[0])
  PST.Str(string("."))
  PST.Dec(DNS_Server[1])
  PST.Str(string("."))
  PST.Dec(DNS_Server[2])
  PST.Str(string("."))
  PST.Dec(DNS_Server[3])
  PST.Str(string(PST#NL))   


  PST.Str(string("MAC Address: "))
  PST.Hex(MAC_Address[0],2)
  PST.Str(string("-"))
  PST.Hex(MAC_Address[1],2)
  PST.Str(string("-"))
  PST.Hex(MAC_Address[2],2)
  PST.Str(string("-"))
  PST.Hex(MAC_Address[3],2)
  PST.Str(string("-"))
  PST.Hex(MAC_Address[4],2)
  PST.Str(string("-"))
  PST.Hex(MAC_Address[5],2)
  PST.Str(string(PST#NL))    
  
  PST.Str(string(PST#NL,PST#NL))  


  ' init the Wiznet 5100 chip
  W5100.StartINDIRECT(W5100_DATA0, W5100_ADDR0, W5100_ADDR1, W5100_CS, W5100_RD, W5100_WR, W5100_RST, W5100_SEN)

  ' setup the address information we got from DHCP
  
  W5100.InitAddresses(true, @MAC_Address[0], @GatewayIP[0], @SubnetMask[0], @IP[0])

  ' open a socket for TCP
  W5100.SocketOpen(0, W5100#_TCPPROTO, listenPort, listenPort, @destIP[0])

  'Wait a moment for the socket to get established
  PauseMSec(250)

  ReadStatus(0)
  
  'Try setting up a listen on the TCP socket
  PST.Str(string("Setting TCP on socket 0, port "))
  PST.dec(listenPort)
  PST.Str(string(" to listening", PST#NL))

  W5100.SocketTCPlisten(0)

  'Wait a moment for the socket to listen
  PauseMSec(250)

  ReadStatus(0)

  'outa[LED] := !outa[LED]

  'Infinite loop of the server
  repeat

    PST.Str(string("Waiting for a client to connect....", PST#NL))
    repeat while !W5100.SocketTCPestablished(0)

    PST.Str(string("connection established...", PST#NL))

    'Initialize the buffers and bring the data over
    bytefill(@data, 0, bytebuffersize)    
    repeat
      packetSize := W5100.rxTCP(0, @data)
    while packetSize == 0  

    PST.Str(string("Packet from browser:", PST#NL))
    PST.Str(@data[0])

    bytefill(@fileName[0], 0, 256)
    
    if @data[0] <> 0
      if STR.startsWithCharacters(STR.tokenizeCharacters(@data[0]), string("GET"))
      
        ' get the file name
        token := STR.tokenizeCharacters(0)
        ' if the file name length is less than 2, assume it's a '/' which means give them the default page
        if STRSIZE(token) < 2
          bytemove(@fileName[0], string("INDEX.HTM"), 9)
          fileName[9] := 0
        else
          ' copy the file name over stripping off the initial '/'
          bytemove(@fileName[0], token+1, STRSIZE(token)-1)

        ' according to docs I read, this is the proper initial response for HTTP/1.1 (HTTP/1.0 doesn't have this) 
        ' doing this causes IE to fail to display the page, so I have it commented out
        'token := STR.tokenizeCharacters(0)
        'if STR.startsWithCharacters(token, string("HTTP/1.1"))
          'StringSend(0, string("HTTP/1.1 100 Continue"))

        ' force upper case filenames (we only have all uppercase file names on the SD card)
        STR.charactersToUpperCase(@fileName[0])
        PST.Str(string("Serving file: "))
        PST.Str(@fileName[0])
        PST.Str(string(PST#NL))
          
        ' see if the file is on the SD card
        token := DoesFileExistOnCard(@fileName[0])

        ' token will be non-zero if the file was found on the SD card  
        if token <> 0
          ' so, open the file
          result := SDCard.openFile(@fileName[0],"r")
           
          fileSize := SDCard.listSize
          if fileSize > 0
            'status line
            StringSend(0, string("HTTP/1.1 200 OK", PST#NL, PST#LF))
           
            'header fields
            StringSend(0, string("Server: Parallax Spinneret Web Server", PST#NL, PST#LF))
            ' this tells the client we don't support persistent connections
            StringSend(0, string("Connection: close", PST#NL, PST#LF))

            ' pick a type base on the file extension
            if STR.endsWithCharacters(@fileName[0], string(".HTM"))
              StringSend(0, string("Content-Type: text/html", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".TXT"))
              StringSend(0, string("Content-Type: text/plain", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".PDF"))
              StringSend(0, string("Content-Type: application/pdf", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".ZIP"))
              StringSend(0, string("Content-Type: application/zip", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".JPG"))
              StringSend(0, string("Content-Type: image/jpeg", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".PNG"))
              StringSend(0, string("Content-Type: image/png", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".GIF"))
              StringSend(0, string("Content-Type: image/gif", PST#NL, PST#LF))
            elseif STR.endsWithCharacters(@fileName[0], string(".ICO"))
              StringSend(0, string("Content-Type: image/vnd.microsoft.icon", PST#NL, PST#LF))

            'this blank line is required, it marks the end of the header fields  
            StringSend(0, string(PST#NL, PST#LF))
           
            if fileSize < maxPacketSendSize
              ' send the file in one packet
              PST.Str(string("sending small file", PST#NL)) 
              SDCard.readData(@data[0], fileSize)
              W5100.txTCP(0, @data[0], fileSize)
            else
              ' send the file in a bunch of packets
              PST.Str(string("sending large file")) 
              repeat
                SDCard.readData(@data[0], maxPacketSendSize)
                W5100.txTCP(0, @data[0], maxPacketSendSize)
                PST.Str(string("*"))
                fileSize -= maxPacketSendSize
                ' once the remaining fileSize is less then the max packet size, just send that remaining bit and quit the loop
                if fileSize < maxPacketSendSize and fileSize > 0
                  SDCard.readData(@data[0], fileSize)
                  W5100.txTCP(0, @data[0], fileSize)
                  PST.Str(string("."))
                  quit
             PST.Str(string(PST#NL))
  
          SDCard.closeFile
          
        elseif STR.startsWithCharacters(STR.tokenizeCharacters(@filename[0]), String("DIRECTORY")) == TRUE
        
          StringSend(0, string("HTTP/1.1 200 OK", CR, LF))
          'optional header
          StringSend(0, string("Server: Parallax Spinneret Web Server/demo 1", CR))
          StringSend(0, string("Connection: close", CR))
          StringSend(0, string("Content-Type: text/html", CR, LF))
          StringSend(0, string(CR, LF))
          'File
          StringSend(0, string("<HTML>", CR))
          StringSend(0, string("<HEAD>", CR))
          StringSend(0, string("<TITLE>"))
          StringSend(0, string("SD Card Directory Page"))
          StringSend(0, string("</TITLE>", CR))
          StringSend(0, string("</HEAD>", CR))    
          StringSend(0, string("<BODY>", CR))
          StringSend(0, string("<FONT FACE=ARIAL SIZE=4><BR>"))
          StringSend(0, string("SD Card Directory Page<BR><BR>")) 
          StringSend(0, string("<TABLE Border=1>", CR))
          StringSend(0, string("<TR><TD>Filename</TD><TD>Open in Browser</TD><TD>Edit</TD></TR>", CR))
          SDCard.listName("T")
          repeat
            token := nextFindFile
            if token == 0
              quit
            StringSend(0, string("<TR><TD>", CR))
            StringSend(0, token)
            StringSend(0, string("</TD><TD><A Href=",QUOTE))
            StringSend(0, token)
            StringSend(0, string(QUOTE,">Open</A></TD><TD>Edit</TD></TR>",CR)) 
          StringSend(0, string("</TABLE>", CR))
          StringSend(0, string("</FONT>", CR))     
          StringSend(0, string("</BODY>", CR))
          StringSend(0, string("</HTML>", CR))
          StringSend(0, string(CR, LF))         
          PST.Str(String("Served Dynamic SD Card Directory Page", PST#NL, PST#LF))

        elseif STR.startsWithCharacters(STR.tokenizeCharacters(@filename[0]), String("SWITCH")) == TRUE
          outa[XLED] := !outa[XLED]

          PageCount++

          StringSend(0, string("<HTML>", CR))
          StringSend(0, string("<HEAD>", CR))
          StringSend(0, string("<meta http-equiv=refresh content=1;/LED>", CR))
          StringSend(0, string("<TITLE>"))
          StringSend(0, string("Led Switch"))
          StringSend(0, string("</TITLE>", CR))
          StringSend(0, string("</HEAD>", CR))
          StringSend(0, string("<BODY><FONT FACE=ARIAL SIZE=7><BR><BR>", CR))
          StringSend(0, string("<FONT COLOR=ORANGE>", CR))
          StringSend(0, string("&nbsp &nbsp *SWITCH*"))
          StringSend(0, string("<BR><BR></BODY>", CR))
          StringSend(0, string("</HTML>", CR))
          StringSend(0, string(CR, LF))


        elseif STR.startsWithCharacters(STR.tokenizeCharacters(@filename[0]), String("LED")) == TRUE   '= Youy IP + /led

          led_stat := ina[XLED]

          StringSend(0, string("<HTML>", CR))
          StringSend(0, string("<HEAD>", CR))
          StringSend(0, string("<meta http-equiv=refresh content=10>", CR))  ' refrech 10 sec
          StringSend(0, string("<TITLE>"))
          StringSend(0, string("Led Status"))
          StringSend(0, string("</TITLE>", CR))
          StringSend(0, string("</HEAD>", CR))
          StringSend(0, string("<BODY>", CR))
          StringSend(0, string("<FONT FACE=ARIAL SIZE=7><BR><BR>"))
          StringSend(0, string("<li><a href=SWITCH>LED Switch</li><BR>",CR))
          StringSend(0, string("&nbsp &nbsp LED ="))
           if led_stat == 1
            StringSend(0, string("<FONT COLOR=red>", CR))
            StringSend(0, string(" ON ",cr))
           else
            StringSend(0, string("<FONT COLOR=BLACK>", CR))
            StringSend(0, string(" OFF ",cr))

          StringSend(0, string("<FONT COLOR=BLACK><BR><BR>", CR))
          StringSend(0, string("<FONT FACE=ARIAL SIZE=2><BR><BR>"))
          StringSend(0, string("This LED has been served "))
          StringSend(0, string("<b>"))
          StringSend(0, STR.numberToDecimal(PageCount, 5))
          StringSend(0, string("</b>"))
          StringSend(0, string(" times since powering on the module"))
          StringSend(0, string("</FONT>", CR))
          StringSend(0, string("</BODY>", CR))
          StringSend(0, string("</HTML>", CR))
          StringSend(0, string(CR, LF))

        elseif STR.startsWithCharacters(STR.tokenizeCharacters(@filename[0]), String("EDITGET.HTM?")) == TRUE         
          'File
          StringSend(0, string("<HTML>", CR))
          StringSend(0, string("<HEAD>", CR))
          StringSend(0, string("<TITLE>"))
          StringSend(0, string("File Editing"))
          StringSend(0, string("</TITLE>", CR))
          StringSend(0, string("</HEAD>", CR))    
          StringSend(0, string("<BODY>", CR))
          StringSend(0, string("<FONT FACE=ARIAL SIZE=3><BR><BR>"))
          StringSend(0, string("Making a New File??  :)<BR>", CR))
          'SDCard.DeleteEntry(String("TEST.TXT"))
          'SDCard.NewFile(String("TEST.TXT"))
          'SDCard.OpenFile(String("TEST.TXT"),("W"))
          
          bytefill(@EditorFileName[0],0,13)
          Counter := 26       ' "GET /EDITGET.HTM?FILENAME="
          Repeat Until data[Counter] == "&"
            Counter++
            
          bytemove(@EditorFileName[0], @data[26], Counter-26)
          
          if DoesFileExistOnCard(@EditorFileName[0]) == 0
            SDCard.NewFile(@EditorFileName[0])
            StringSend(0,@EditorFileName)
            StringSend(0, string("<BR><BR><BR><BR>Wait Please", CR))  
            PST.DEC(SDCard.OpenFile(@EditorFileName[0],("W")))
            Counter:=Counter+3
            Repeat Until data[Counter] == "&"
              SDCard.WriteByte(data[Counter])  
              Counter++
            SDCard.CloseFile
            StringSend(0, string("<BR><BR>OK It's Closed Now<BR>", CR)) 
            StringSend(0, string("Wrote: ", CR))
            StringSend(0, STR.numberToDecimal(Counter-(26+STRSize(@EditorFileName[0])), 4))    
            StringSend(0, string(" Bytes<BR><BR>", CR))
            StringSend(0, string("Check it out from the <a href=", QUOTE,"directory",QUOTE,">Directory</a>", CR))
          else
            StringSend(0, string("<BR><BR><BR><BR>That File is already on the Card! You need to pick a different filename<br><br>", CR))
            StringSend(0, string("Check the <a href=", QUOTE,"directory",QUOTE,">Directory</a>", CR))
          
          StringSend(0, string("</FONT>", CR))     
          StringSend(0, string("</BODY>", CR))
          StringSend(0, string("</HTML>", CR))
          
        else
        
          StringSend(0, string("HTTP/1.1 200 OK", CR, LF))
           
          'optional header
          StringSend(0, string("Server: Parallax Spinneret Web Server/demo 1", CR))
          StringSend(0, string("Connection: close", CR))
          StringSend(0, string("Content-Type: text/html", CR, LF))
          StringSend(0, string(CR, LF))
           
          'File
          StringSend(0, string("<HTML>", CR))
          StringSend(0, string("<HEAD>", CR))
          StringSend(0, string("<TITLE>"))
          StringSend(0, string("404"))
          StringSend(0, string("</TITLE>", CR))
          StringSend(0, string("</HEAD>", CR))    
          StringSend(0, string("<BODY>", CR))
          StringSend(0, string("<FONT FACE=ARIAL SIZE=7><BR><BR>"))
          StringSend(0, string("404 The File Requested Can Not be Found on this server", CR))
          StringSend(0, string("</FONT>", CR))     
          StringSend(0, string("</BODY>", CR))
          StringSend(0, string("</HTML>", CR))
          StringSend(0, string(CR, LF))
              
          PST.Str(string("File Not Found: "))
          PST.Str(@fileName[0])
          PST.Str(string(PST#NL))
          
          SDCard.closeFile 

    ' we don't support persistent connections, so disconnect here     
    W5100.SocketTCPdisconnect(0)
    PauseMSec(25)
    
    'Connection terminated
    W5100.SocketClose(0)
    PST.Str(string("Connection complete", PST#NL, PST#NL))

    'Once the connection is closed, need to open socket again
    OpenSocketAgain(0)

PUB DoesFileExistOnCard(filenameptr)| token

  SDCard.listName("T")
  repeat
    token := nextFindFile
    if token == 0
      quit
    if strcomp(filenameptr, token)
      quit
  return token

PUB nextFindFile | temp, index  
  temp := SDCard.listName(" ")
  repeat index from 0 to 11
    if byte[temp][index] == 32
      byte[temp][index] := 0
      quit
  return temp

PRI StringSend(socket, _dataPtr)
  W5100.txTCP(socket, _dataPtr, strsize(_dataPtr))

PRI OpenSocketAgain(socket)
  W5100.SocketOpen(socket, W5100#_TCPPROTO, listenPort, listenPort, @destIP[0])
  W5100.SocketTCPlisten(socket)

PRI ReadStatus(socket) | status
  PST.Str(string("Socket "))
  PST.Dec(socket)
  PST.STR(string(" Status Register: "))
  W5100.readIND((W5100#_S0_SR + (socket * $0100)), @status, 1)

  case status
    W5100#_SOCK_CLOSED : PST.Str(string("$00 - socket closed", PST#NL, PST#NL))
    W5100#_SOCK_INIT   : PST.Str(string("$13 - socket initalized", PST#NL, PST#NL))
    W5100#_SOCK_LISTEN : PST.Str(string("$14 - socket listening", PST#NL, PST#NL))
    W5100#_SOCK_ESTAB  : PST.Str(string("$17 - socket established", PST#NL, PST#NL))    
    W5100#_SOCK_UDP    : PST.Str(string("$22 - socket UDP open", PST#NL, PST#NL))

  return status
  
PRI PauseMSec(Duration)
'***************************************
''  Pause execution for specified milliseconds.
''  This routine is based on the set clock frequency.
''  
''  params:  Duration = number of milliseconds to delay                                                                                               
''  return:  none
  
  waitcnt(((clkfreq / 1_000 * Duration - 3932) #> 381) + cnt)
